from django.core.exceptions import ObjectDoesNotExist
from edc_sites import site_sites
from edc_sites.utils import add_or_update_django_sites, get_site_model_cls
from edc_test_utils.get_user_for_tests import get_user_for_tests
from edc_visit_schedule.models import VisitSchedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from subject_dashboard_app.sites import all_sites
from subject_dashboard_app.visit_schedule import visit_schedule1


class TestCaseMixin:
    @classmethod
    def setUpTestData(cls):
        for obj in get_site_model_cls().objects.all():
            try:
                obj.siteprofile.delete()
            except ObjectDoesNotExist:
                pass
            obj.delete()
        get_site_model_cls().objects.all().delete()
        site_sites._registry = {}
        site_sites.register(*all_sites)
        add_or_update_django_sites(verbose=False)
        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule1)
        VisitSchedule.objects.update(active=False)
        site_visit_schedules.to_model(VisitSchedule)
        cls.user = get_user_for_tests()



from __future__ import annotations

from typing import Any

from django.core.exceptions import ImproperlyConfigured
from edc_action_item.view_mixins import ActionItemViewMixin
from edc_appointment.view_mixins import AppointmentViewMixin
from edc_consent.view_mixins import ConsentViewMixin
from edc_dashboard.view_mixins import EdcViewMixin
from edc_dashboard.views import DashboardView
from edc_data_manager.view_mixins import DataManagerViewMixin
from edc_locator.view_mixins import SubjectLocatorViewMixin
from edc_metadata.view_mixins import MetadataViewMixin
from edc_navbar.view_mixin import NavbarViewMixin
from edc_sites.site import sites
from edc_visit_schedule.view_mixins import VisitScheduleViewMixin

from ..view_mixins import RegisteredSubjectViewMixin, SubjectVisitViewMixin


class VerifyRequisitionMixin:
    def get_context_data(self, **kwargs) -> dict[str, Any]:
        scanning = self.kwargs.get("scanning")
        kwargs.update(scanning=scanning)
        return super().get_context_data(**kwargs)


class SubjectDashboardView(
    EdcViewMixin,
    NavbarViewMixin,
    VisitScheduleViewMixin,
    MetadataViewMixin,
    ConsentViewMixin,
    SubjectLocatorViewMixin,
    ActionItemViewMixin,
    DataManagerViewMixin,
    SubjectVisitViewMixin,
    AppointmentViewMixin,
    RegisteredSubjectViewMixin,
    VerifyRequisitionMixin,
    DashboardView,
):
    navbar_selected_item = "consented_subject"

    dashboard_url_name = "subject_dashboard_url"
    dashboard_template = "subject_dashboard_template"

    default_manager = "on_site"

    def __init__(self, **kwargs):
        if not self.navbar_name:
            raise ImproperlyConfigured(f"'navbar_name' cannot be None. See {repr(self)}.")
        super().__init__(**kwargs)

    @property
    def manager(self) -> str:
        """Returns the name of the model manager"""
        if sites.user_may_view_other_sites(self.request):
            return "objects"
        return self.default_manager



from django.contrib import messages
from django.http.response import HttpResponseRedirect
from django.urls.base import reverse
from edc_appointment.models import Appointment
from edc_constants.constants import YES
from edc_lab.model_mixins import RequisitionModelMixin
from edc_lab.models import Consignee
from edc_label.job_result import JobResult
from edc_label.printers_mixin import PrinterError, PrintServerError
from edc_metadata.constants import KEYED, REQUIRED
from edc_metadata.models import RequisitionMetadata

from ..requisition_labels import RequisitionLabels
from ..requisition_report import RequisitionReport
from .base_requisition_view import BaseRequisitionView


class RequisitionPrintActionsView(BaseRequisitionView):
    job_result_cls = JobResult
    requisition_report_cls = RequisitionReport
    requisition_labels_cls = RequisitionLabels
    print_selected_button = "print_selected_labels"
    print_all_button = "print_all_labels"
    print_requisition = "print_requisition"
    checkbox_name = "selected_panel_names"

    def __init__(self, **kwargs):
        self._appointment = None
        self._selected_panel_names = []
        self._requisition_metadata = None
        self._requisition_model_cls = None
        self.consignee = None
        super().__init__(**kwargs)

    def post(self, request, *args, **kwargs):
        response = None
        if self.selected_panel_names:
            if self.request.POST.get("submit") in [
                self.print_all_button,
                self.print_selected_button,
            ]:
                self.print_labels_action()
            elif self.request.POST.get("submit_print_requisition"):
                self.consignee = Consignee.objects.get(
                    pk=self.request.POST.get("submit_print_requisition")
                )
                response = self.render_manifest_to_response_action()
        if not response:
            subject_identifier = request.POST.get("subject_identifier")
            success_url = reverse(
                self.get_success_url(),
                kwargs=dict(
                    subject_identifier=subject_identifier,
                    appointment=str(self.appointment.pk),
                ),
            )
            response = HttpResponseRedirect(redirect_to=f"{success_url}")
        return response

    def print_labels_action(self):
        labels = self.requisition_labels_cls(
            requisition_metadata=self.requisition_metadata.filter(
                panel_name__in=self.selected_panel_names
            ),
            panel_names=self.selected_panel_names,
            appointment=self.appointment,
            user=self.request.user,
        )

        if labels.zpl_data:
            try:
                job_id = self.clinic_label_printer.stream_print(zpl_data=labels.zpl_data)
            except (PrintServerError, PrinterError) as e:
                messages.error(self.request, str(e))
            else:
                job_result = self.job_result_cls(
                    name=labels.label_template_name,
                    copies=1,
                    job_ids=[job_id],
                    printer=self.clinic_label_printer,
                )
                messages.success(self.request, job_result.message)
        if labels.requisitions_not_printed:
            panels = ", ".join([str(r.panel_object) for r in labels.requisitions_not_printed])
            messages.warning(
                self.request, f"Some selected labels were not printed. See {panels}."
            )

    def render_manifest_to_response_action(self):
        panel_names = [r.panel.name for r in self.verified_requisitions]
        if panel_names:
            requisition_report = self.requisition_report_cls(
                appointment=self.appointment,
                selected_panel_names=panel_names,
                consignee=self.consignee,
                request=self.request,
            )
            response = requisition_report.render_to_response()
        else:
            messages.error(self.request, 'Nothing to do. No "verified" requisitions selected.')
            response = None
        return response

    @property
    def requisition_metadata(self):
        """Returns a queryset of keyed or required RequisitionMetadata for this
        appointment.
        """
        if not self._requisition_metadata:
            appointment = Appointment.objects.get(pk=self.request.POST.get("appointment"))
            subject_identifier = self.request.POST.get("subject_identifier")
            opts = dict(
                subject_identifier=subject_identifier,
                visit_schedule_name=appointment.visit_schedule_name,
                schedule_name=appointment.schedule_name,
                visit_code=appointment.visit_code,
                visit_code_sequence=appointment.visit_code_sequence,
            )
            self._requisition_metadata = RequisitionMetadata.objects.filter(
                entry_status__in=[KEYED, REQUIRED], **opts
            )
        return self._requisition_metadata

    @property
    def selected_panel_names(self):
        """Returns a list of panel names selected on the page.

        Returns all on the page if "print all" is submitted.
        """
        if not self._selected_panel_names:
            if self.request.POST.get("submit") == self.print_all_button:
                for metadata in self.requisition_metadata:
                    self._selected_panel_names.append(metadata.panel_name)
            else:
                self._selected_panel_names = (
                    self.request.POST.getlist(self.checkbox_name) or []
                )
        return self._selected_panel_names

    @property
    def appointment(self):
        if not self._appointment:
            self._appointment = Appointment.objects.get(
                pk=self.request.POST.get("appointment")
            )
        return self._appointment

    @property
    def requisition_model_cls(self):
        if not self._requisition_model_cls:
            for v in self.appointment.visit_model_cls().__dict__.values():
                try:
                    model_cls = getattr(getattr(v, "rel"), "related_model")
                except AttributeError:
                    pass
                else:
                    if issubclass(model_cls, RequisitionModelMixin):
                        self._requisition_model_cls = model_cls
        return self._requisition_model_cls

    @property
    def verified_requisitions(self):
        """Returns a list of "verified" requisition model instances related
        to this appointment.
        """
        verified_requisitions = []
        for k, v in self.appointment.visit_model_cls().__dict__.items():
            try:
                model_cls = getattr(getattr(v, "rel"), "related_model")
            except AttributeError:
                pass
            else:
                if issubclass(model_cls, RequisitionModelMixin):
                    verified_requisitions.extend(
                        list(
                            getattr(self.appointment.related_visit, k).filter(
                                clinic_verified=YES,
                                panel__name__in=self.selected_panel_names,
                            )
                        )
                    )
        return verified_requisitions



from .refresh_appointments_view import RefreshAppointmentsView
from .requisition_print_actions_view import RequisitionPrintActionsView
from .requisition_verify_actions_view import RequisitionVerifyActionsView
from .subject_dashboard_view import SubjectDashboardView



from tempfile import mkdtemp

from django.contrib import messages
from edc_constants.constants import YES
from edc_identifier.utils import convert_to_human_readable
from edc_lab.model_mixins import RequisitionModelMixin
from edc_lab.models.manifest.shipper import Shipper
from edc_pdf_reports import Report
from edc_utils import get_utcnow
from reportlab.graphics.barcode import code39
from reportlab.lib import colors
from reportlab.lib.units import cm, mm
from reportlab.platypus import Paragraph, Spacer, Table, TableStyle


class RequisitionReport(Report):
    def __init__(
        self,
        appointment=None,
        selected_panel_names=None,
        consignee=None,
        request=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._requisitions = []
        self.request = request
        self.selected_panel_names = selected_panel_names
        self.site = request.site
        self.appointment = appointment
        self.user = request.user
        self.consignee = consignee
        self.contact_name = f"{self.user.first_name} {self.user.last_name}"
        self.image_folder = mkdtemp()
        self.timestamp = get_utcnow().strftime("%Y%m%d%H%M%S")
        self.report_filename = f"requisition_{self.timestamp}.pdf"

    @property
    def shipper(self):
        """Return a shipper model instance."""
        try:
            shipper = Shipper.objects.all()[0]
        except IndexError:
            messages.error(self.request, "Unable to print report. Please define the shipper.")
            shipper = Shipper()
        return shipper

    def get_report_story(self, **kwargs):
        story = [
            Paragraph(
                "PATIENT SPECIMEN REQUISITION / MANIFEST",
                self.styles["line_label_center"],
            )
        ]

        data = [
            [
                Paragraph("DATE", self.styles["line_label"]),
                Paragraph("REFERENCE", self.styles["line_label"]),
            ],
            [
                Paragraph(get_utcnow().strftime("%Y-%m-%d"), self.styles["line_data_largest"]),
                Paragraph(
                    convert_to_human_readable(self.timestamp),
                    self.styles["line_data_largest"],
                ),
            ],
        ]
        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data = [
            [
                Paragraph("FROM (complete name and address)", self.styles["line_label"]),
                Paragraph("DELIVER TO (complete name and address)", self.styles["line_label"]),
            ],
            [
                Paragraph(
                    self.formatted_address(**self.shipper_data),
                    self.styles["line_data_large"],
                ),
                Paragraph(
                    self.formatted_address(**self.consignee_data),
                    self.styles["line_data_large"],
                ),
            ],
        ]

        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data = [
            [
                Paragraph("CLINIC NAME", self.styles["line_label"]),
                Paragraph("INSTRUCTIONS<br />", self.styles["line_label"]),
            ],
            [
                Paragraph(self.site.name.title(), self.styles["line_data_large"]),
                Paragraph(self.description or "", self.styles["line_data"]),
            ],
            [Paragraph("TEL/MOBILE/FAX", self.styles["line_label"]), ""],
            [
                Paragraph(
                    f'T:{self.shipper.telephone or "?"} M:{self.shipper.mobile or "?"} '
                    f'F:{self.shipper.fax or "?"}',
                    self.styles["line_data_large"],
                ),
                "",
            ],
            [Paragraph("EMAIL", self.styles["line_label"]), ""],
            [Paragraph(self.shipper.email or "", self.styles["line_data_large"]), ""],
        ]
        t1 = Table(data)
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 1), (0, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (0, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("INNERGRID", (0, 2), (1, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (1, 3), 0.25, colors.black),
                    ("INNERGRID", (0, 4), (1, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 5), (1, 5), 0.25, colors.black),
                    ("SPAN", (1, 1), (1, 5)),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t1)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Table(
                [
                    [
                        Paragraph(
                            "I DECLARE THE INFORMATION TO BE TRUE AND CORRECT",
                            self.styles["line_label"],
                        )
                    ]
                ]
            )
        )

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        data = [
            [
                Paragraph(self.contact_name, self.styles["line_data_large"]),
                "",
                Paragraph(
                    get_utcnow().strftime("%Y-%m-%d %H:%M"),
                    self.styles["line_data_large"],
                ),
            ],
            [
                Paragraph("SIGNATURE (sign next to your name.)", self.styles["line_label"]),
                "",
                Paragraph("DATE", self.styles["line_label"]),
            ],
        ]

        t1 = Table(data, colWidths=(None, 2 * cm, None))
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (2, 0), (2, 1), 0.25, colors.black),
                ]
            )
        )

        story.append(t1)
        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Paragraph(
                "Place a check-mark in the right-most column for each specimen received.",
                self.styles["line_label"],
            )
        )

        story = self.append_requisition_items_story(story)

        story.append(Spacer(0.1 * cm, 1 * cm))

        story.append(Paragraph("Comments (if any):", self.styles["line_label"]))
        story.append(Spacer(0.1 * cm, 2 * cm))

        story.append(
            Paragraph(
                "The above items marked with a check have been received in good order:",
                self.styles["line_label"],
            )
        )

        story.append(Spacer(0.1 * cm, 1 * cm))

        data = [
            [
                Paragraph("", self.styles["line_data_large"]),
                "",
                Paragraph("", self.styles["line_data_large"]),
            ],
            [
                Paragraph(
                    "SIGNATURE OF LABORATORY TECHNICIAN (Write name and sign.)",
                    self.styles["line_label"],
                ),
                "",
                Paragraph("DATE", self.styles["line_label"]),
            ],
        ]

        t1 = Table(data, colWidths=(None, 2 * cm, None))
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (2, 0), (2, 1), 0.25, colors.black),
                ]
            )
        )
        story.append(t1)

        return story

    @property
    def requisitions(self):
        """Returns a list of requisition model instances related
        to this appointment.
        """
        if not self._requisitions:
            for k, v in self.appointment.related_visit_model_cls().__dict__.items():
                try:
                    model_cls = getattr(getattr(v, "rel"), "related_model")
                except AttributeError:
                    pass
                else:
                    if issubclass(model_cls, RequisitionModelMixin):
                        self._requisitions.extend(
                            list(
                                getattr(self.appointment.related_visit, k).filter(
                                    clinic_verified=YES,
                                    panel__name__in=self.selected_panel_names,
                                )
                            )
                        )
        return self._requisitions

    @property
    def shipper_data(self):
        data = self.shipper.__dict__
        if self.contact_name.strip():
            data.update(contact_name=self.contact_name)
        return data

    @property
    def consignee_data(self):
        data = self.consignee.__dict__
        return data

    @staticmethod
    def formatted_address(**kwargs):
        data = {
            "contact_name": None,
            "name": None,
            "address": None,
            "city": None,
            "state": None,
            "postal_code": "0000",
            "country": None,
        }
        data.update(**kwargs)
        data_list = [
            v
            for v in [
                data.get("contact_name"),
                data.get("name"),
                data.get("address"),
                (
                    data.get("city")
                    if not data.get("state")
                    else "{} {}".format(data.get("city"), data.get("state"))
                ),
                data.get("postal_code"),
                data.get("country"),
            ]
            if v
        ]
        return "<br />".join(data_list)

    @property
    def description(self):
        return (
            "Please process the following specimens according to the "
            "panels listed below. Use the specimen identifier and "
            "subject identifier as a reference for panel results."
        )

    def append_requisition_items_story(self, story):
        story.append(Spacer(0.1 * cm, 0.5 * cm))
        index = 0
        for requisition in self.requisitions:
            table_data = []
            if index == 0:
                table_data.append(
                    [
                        Paragraph("BARCODE", self.styles["line_label_center"]),
                        Paragraph("ITEM", self.styles["line_label_center"]),
                        Paragraph("SPECIMEN", self.styles["line_label_center"]),
                        Paragraph("SUBJECT", self.styles["line_label_center"]),
                        Paragraph("PANEL", self.styles["line_label_center"]),
                        Paragraph("TYPE", self.styles["line_label_center"]),
                        Paragraph("DRAWN", self.styles["line_label_center"]),
                        Paragraph("RECV'D", self.styles["line_label_center"]),
                    ]
                )
            barcode = code39.Standard39(
                requisition.requisition_identifier, barHeight=10 * mm, stop=1
            )
            item_count = requisition.item_count or 1
            for count in range(1, item_count + 1):
                table_data.append(
                    [
                        barcode,
                        Paragraph(str(index + 1), self.styles["row_data"]),
                        Paragraph(
                            f"{requisition.human_readable_identifier} "
                            f"({count}/{item_count})",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            requisition.subject_identifier or "?",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            f"{requisition.panel_object.verbose_name} <br />"
                            f"({requisition.panel_object.abbreviation})",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            f"{requisition.panel_object.aliquot_type}",
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            requisition.drawn_datetime.strftime("%Y-%m-%d"),
                            self.styles["row_data"],
                        ),
                        Paragraph("", self.styles["row_data"]),
                    ]
                )
                index += 1

            t1 = Table(table_data)
            t1.setStyle(
                TableStyle(
                    [
                        ("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black),
                        ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                        ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ]
                )
            )
            story.append(t1)

        return story



from django.test import TestCase
from edc_appointment.models import Appointment
from edc_utils import get_utcnow

from subject_dashboard_app.models import BadSubjectVisit, SubjectVisit, TestModel


class DummyModelWrapper:
    def __init__(self, **kwargs):
        pass


class TestPrintManifest(TestCase):
    def setUp(self):
        self.appointment = Appointment.objects.create(
            visit_code="1000", appt_datetime=get_utcnow()
        )
        self.subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, subject_identifier="12345"
        )
        self.bad_subject_visit = BadSubjectVisit.objects.create(
            appointment=self.appointment, subject_identifier="12345"
        )
        self.test_model = TestModel.objects.create(subject_visit=self.subject_visit)



from edc_auth.constants import (
    AUDITOR_ROLE,
    CLINICIAN_ROLE,
    CLINICIAN_SUPER_ROLE,
    NURSE_ROLE,
)
from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

site_auths.add_post_update_func(
    "edc_subject_dashboard", remove_default_model_permissions_from_edc_permissions
)

SUBJECT_VIEW = "SUBJECT_VIEW"

site_auths.add_custom_permissions_tuples(
    model="edc_subject_dashboard.edcpermissions",
    codename_tuples=(
        ("edc_subject_dashboard.view_subject_listboard", "Can access subject listboard"),
        ("edc_subject_dashboard.nav_subject_section", "Can access nav_subject_section"),
    ),
)

site_auths.add_group(
    "edc_subject_dashboard.view_subject_listboard",
    "edc_subject_dashboard.nav_subject_section",
    name=SUBJECT_VIEW,
)

site_auths.update_role(SUBJECT_VIEW, name=CLINICIAN_ROLE)
site_auths.update_role(SUBJECT_VIEW, name=NURSE_ROLE)
site_auths.update_role(SUBJECT_VIEW, name=CLINICIAN_SUPER_ROLE)
site_auths.update_role(SUBJECT_VIEW, name=AUDITOR_ROLE)
